@@ -183,6 +183,8 @@ class Speciation(object):
             "logfO2" : ("", ""),
             "cal/kg.H2O" : ("energy supply", "cal/kg H2O"),
             "Log ion-H+ activity ratio" : ("Log ion-H+ activity ratio", ""),
+            "log_fugacity" : ("log fugacity", "log(bar)"),
+            "fugacity" : ("fugacity", "bar"),
         }
         
         out = unit_name_dict.get(subheader)
@@ -243,6 +245,9 @@ class Speciation(object):
         elif col_data.columns.get_level_values(1) == 'log_gamma':
             y = [10**float(s[0]) if s[0] != 'NA' else float("nan") for s in col_data.values.tolist()]
             out_unit = 'gamma'
+        elif col_data.columns.get_level_values(1) == 'log_fugacity':
+            y = [10**float(s[0]) if s[0] != 'NA' else float("nan") for s in col_data.values.tolist()]
+            out_unit = 'fugacity'
         else:
             y = [float(s[0]) if s[0] != 'NA' else float("nan") for s in col_data.values.tolist()]
             out_unit = col_data.columns.get_level_values(1)[0]
@@ -339,7 +344,7 @@ class Speciation(object):
         plt.show()
     
 
-    def barplot(self, y, yrange=None, show_trace=True,
+    def barplot(self, y, yrange=None, convert_log=True, show_trace=True,
                 show_legend=True, show_missing=True, legend_loc="best",
                 colormap="viridis", bg_color="white", save_as=None):
         
@@ -354,6 +359,11 @@ class Speciation(object):
        
         yrange : list of numeric, optional
             Sets the lower and upper limits of the y axis.
+            
+        convert_log : bool, default True
+            Convert units "log_activity", "log_molality", "log_gamma", and
+            "log_fugacity" to "activity", "molality", "gamma", and "fugacity",
+            respectively?
         
         show_trace : bool, default True
             Show asterisks for columns with numerical values but are too short
@@ -425,8 +435,8 @@ class Speciation(object):
                        "'{}' are non-numeric and cannot be plotted.".format(y_col.columns.get_level_values(0)[0]))
                 raise Exception(msg)
             
-            if [abs(y0) for y0 in y_vals] != y_vals: # convert to bar-friendly units if possible
-                if subheader in ["log_activity", "log_molality", "log_gamma"]:
+            if convert_log and [abs(y0) for y0 in y_vals] != y_vals: # convert to bar-friendly units if possible
+                if subheader in ["log_activity", "log_molality", "log_gamma", "log_fugacity"]:
                     y_plot, out_unit = self.__convert_aq_units_to_log_friendly(yi, rows=x)
                     unit_type, unit = self.__get_unit_info(out_unit)
                 else:
@@ -1420,7 +1430,7 @@ class AqEquil():
     def speciate(self,
                  input_filename,
                  db="wrm",
-                 redox_flag=0,
+                 redox_flag="logfO2",
                  redox_aux="Fe+3",
                  default_logfO2=-6,
                  exclude=[],
@@ -1438,6 +1448,7 @@ class AqEquil():
                  get_redox=True,
                  redox_type="Eh",
                  get_ion_activity_ratios=True,
+                 get_fugacity=True,
                  get_affinity_energy=False,
                  rxn_filename=None,
                  not_limiting=["H+", "OH-", "H2O"],
@@ -1504,37 +1515,28 @@ class AqEquil():
             data1 files are already compiled by EQPT, while data0 files will be
             automatically compiled for you if `custom_db` is True.
         
-        redox_flag : int, default 0
-            Values corresponding to redox options in the EQ3/6 v8.0 software.
-            For more information see the 'Redox Option' section of the EQ3/6
-            version 8.0 software user's manual. Set sample redox state with the
-            following options:
+        redox_flag : str, default "O2(g)"
+            Determines which column in the sample input file sets the overall
+            redox state of the samples. Options for redox_flag include 'O2(g)',
+            'pe', 'Eh', 'logfO2', and 'redox aux'. The code will search your
+            sample spreadsheet file (see `filename`) for a column corresponding
+            to the option you chose:
             
-            * -3 for O2(g)
-            * -2 for pe (in pe units)
-            * -1 for Eh (volts)
-            *  0 for logfO2 (log bars), or dissolved O2 (see below)
-            *  1 for defining a redox couple (see `redox_aux`)
-             
-            Note that if you are importing water chemistry data from a
-            spreadsheet, a column must be supplied with data that corresponds to
-            the redox option you chose. The column name is important:
-            
-            * -3 must have a column named: O2(g)
-            * -2 must have a column named: pe
-            * -1 must have a column named: Eh
-            *  0 must have a column named: logfO2
-            *  1 must have a column corresponding to the auxilliary basis species
-              selected to form a redox couple with its linked species (see
-              `redox_aux`). For example, the redox couple Fe+2/Fe+3 would need
-              a column named: Fe+3
+            * 'O2(g)' with a valid subheader for a gas
+            * 'pe' with subheader pe
+            * 'Eh' with subheader volts
+            * 'logfO2' with subheader logfO2
+            * 'redox aux' will search for a column corresponding to the
+              auxilliary basis species selected to form a redox couple with its
+              linked strict basis species (see `redox_aux`). For example, the
+              redox couple Fe+2/Fe+3 would require a column named Fe+3
             
             If an appropriate header or redox data cannot be found to define
             redox state, `default_logfO2` is used to set sample logfO2.
             
             There is a special case where dissolved oxygen can be used to impose
-            sample redox state if `redox_flag` is set to 0 and a column named
-            logfO2 does not appear in the sample data sheet. If there is a
+            sample redox state if `redox_flag` is set to logfO2 and a column named
+            logfO2 does not appear in your sample spreadsheet. If there is a
             column corresponding to dissolved oxygen measurements, logfO2 is
             calculated from the equilibrium reaction O2(aq) = O2(g) at the
             temperature and pressure of the sample using the revised Helgeson-
@@ -1619,6 +1621,9 @@ class AqEquil():
         get_ion_activity_ratios : bool, default True
             Calculate ion/H+ activity ratios and neutral species activities?
         
+        get_fugacity : bool, default True
+            Calculate gas fugacities?
+        
         get_affinity_energy : bool, default False
             Calculate affinities and energy supplies of reactions listed in a
             separate user-supplied file?
@@ -1670,6 +1675,20 @@ class AqEquil():
         self._check_sample_input_file(input_filename, exclude, db, custom_db,
                                       charge_balance_on, suppress_missing)
         
+        if redox_flag == "O2(g)" or redox_flag == -3:
+            redox_flag = -3
+        elif redox_flag == "pe" or redox_flag == -2:
+            redox_flag = -2
+        elif redox_flag == "Eh" or redox_flag == -1:
+            redox_flag = -1
+        elif redox_flag == "logfO2" or redox_flag == 0:
+            redox_flag = 0
+        elif redox_flag == "redox aux" or redox_flag == 1:
+            redox_flag = 1
+        else:
+            raise Exception("Unrecognized redox flag. Valid options are 'O2(g)'"
+                            ", 'pe', 'Eh', 'logfO2', 'redox aux'")
+            
         # handle batch_3o naming
         if batch_3o_filename != None:
             if ".rds" in batch_3o_filename[-4:]:
@@ -1691,6 +1710,42 @@ class AqEquil():
 
             self.runeqpt(db, extra_eqpt_output)
             os.environ['EQ36DA'] = os.getcwd()
+            
+            data0_path = "data0." + db
+            
+        else:
+            data0_path = self.eq36da + "/data0." + db
+            
+        if os.path.exists(data0_path) and os.path.isfile(data0_path):
+            with open(data0_path) as data0:
+                data0_lines = data0.readlines()
+                start_index = [i+1 for i, s in enumerate(data0_lines) if s == 'temperatures\n']
+                end_index = [i for i, s in enumerate(data0_lines) if s == 'debye huckel a (adh)\n']
+                db_grids_unformatted = [i.split("pressures")[0] for i in data0_lines[start_index[0]:end_index[0]]]
+                db_grids = [" ".join(i.split()) for i in db_grids_unformatted if i != '']
+                grid_temp = db_grids[0] + " " + db_grids[1]
+                grid_press = db_grids[2] + " " + db_grids[3]
+                grid_temp = grid_temp.split(" ")
+                grid_press = grid_press.split(" ")
+                try:
+                    water_model = data0_lines[1].split("model: ")[1] # extract water model from the second line of data0 file
+                    water_model = water_model.replace("\n", "")
+                except:
+                    water_model = "SUPCRT92"
+#                     print("Water model could not be referenced from {}".format(data0_path)+""
+#                           ". Defaulting to SUPCRT92 water model...")
+                
+                if(water_model not in ["SUPCRT92", "IAPWS95", "DEW"]):
+                    water_model = "SUPCRT92" # the default for EQ3/6
+                    print("Water model given in {}".format(data0_path)+" was not "
+                          "recognized. Defaulting to SUPCRT92 water model...")
+            
+        else: # if a data0 file can't be found, assume default water model, 0-350 C and PSAT
+            water_model = "SUPCRT92"
+            grid_temp = ["0.0100", "50.0000", "100.0000", "150.0000",
+                         "200.0000", "250.0000", "300.0000", "350.0000"]
+            grid_press = ["1.0000", "1.0000", "1.0132", "4.7572",
+                          "15.5365", "39.7365", "85.8378", "165.2113"]
 
         if get_affinity_energy:
             if rxn_filename == None:
@@ -1701,7 +1756,7 @@ class AqEquil():
                 pass
         else:
             rxn_filename = ""
-
+            
         # preprocess for EQ3 using R scripts
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter("always")
@@ -1712,11 +1767,15 @@ class AqEquil():
                                                  exclude=convert_to_RVector(
                                                      exclude),
                                                  redox_flag=redox_flag,
+                                                 redox_aux=redox_aux,
                                                  default_logfO2=default_logfO2,
                                                  charge_balance_on=charge_balance_on,
                                                  suppress_missing=suppress_missing,
                                                  suppress=convert_to_RVector(
                                                      suppress),
+                                                 water_model=water_model,
+                                                 grid_temp=convert_to_RVector(grid_temp),
+                                                 grid_press=convert_to_RVector(grid_press),
                                                  verbose=self.verbose)
 
         for warning in w:
@@ -1768,6 +1827,7 @@ class AqEquil():
                 redox_type=redox_type,
                 get_charge_balance=get_charge_balance,
                 get_ion_activity_ratios=get_ion_activity_ratios,
+                get_fugacity=get_fugacity,
                 get_affinity_energy=get_affinity_energy,
                 not_limiting=convert_to_RVector(not_limiting),
                 batch_3o_filename=batch_3o_filename,
@@ -1793,7 +1853,8 @@ class AqEquil():
         # handle headers and subheaders of input section
         headers = [col.split("_")[0] for col in list(df_input.columns)]
         headers = ["pH" if header == "H+" else header for header in headers]
-        headers = [header+"_(input)" if header not in ["Temperature", "pH", "logfO2"]+exclude else header for header in headers]
+        headers = [header+"_(input)" if header not in ["Temperature", "logfO2"]+exclude else header for header in headers]
+        report_divs[0] = convert_to_RVector(headers) # modify headers in the 'input' section, report_divs[0]
         subheaders = [subheader[1] if len(subheader) > 1 else "" for subheader in [
             col.split("_") for col in list(df_input.columns)]]
         multicolumns = pd.MultiIndex.from_arrays(
@@ -1807,12 +1868,21 @@ class AqEquil():
             df_aq_distribution = df_report[aq_distribution_cols]
             df_aq_distribution = df_aq_distribution.apply(pd.to_numeric, errors='coerce')
 
+            # create a pH column from H+
+            df_aq_distribution["pH"] = -df_aq_distribution["H+"]
+            
             # handle headers of aq_distribution section
             headers = df_aq_distribution.columns
-            subheaders = [aq_dist_type]*len(headers)
+            subheaders = [aq_dist_type]*(len(headers)-1) # -1 because the last column will have subheader pH (see next line)
+            subheaders = subheaders + ["pH"]
             multicolumns = pd.MultiIndex.from_arrays(
                 [headers, subheaders], names=['Sample', ''])
             df_aq_distribution.columns = multicolumns
+            
+            # ensure final pH column is included in report_divs aq_distribution section
+            aq_dist_indx = report_divs.names.index("aq_distribution")
+            report_divs[aq_dist_indx] = convert_to_RVector(list(headers))
+            
             df_join = df_join.join(df_aq_distribution)
 
         if get_mineral_sat:
@@ -1887,7 +1957,20 @@ class AqEquil():
                 [headers, subheaders], names=['Sample', ''])
             df_ion_activity_ratios.columns = multicolumns
             df_join = df_join.join(df_ion_activity_ratios)
-
+            
+        if get_fugacity:
+            fugacity_cols = list(report_divs.rx2('fugacity'))
+            df_fugacity = df_report[fugacity_cols]
+            df_fugacity = df_fugacity.apply(pd.to_numeric, errors='coerce')
+            
+            # handle headers of fugacity section
+            headers = df_fugacity.columns
+            subheaders = ["log_fugacity"]*len(headers)
+            multicolumns = pd.MultiIndex.from_arrays(
+                [headers, subheaders], names=['Sample', ''])
+            df_fugacity.columns = multicolumns
+            df_join = df_join.join(df_fugacity)
+            
         if get_affinity_energy:
             affinity_cols = list(report_divs.rx2('affinity'))
             energy_cols = list(report_divs.rx2('energy'))
@@ -1965,6 +2048,15 @@ class AqEquil():
                 except:
                     dict_sample_data['ion_activity_ratios'] = None
             
+            if get_fugacity:
+                dict_sample_data.update(
+                    {"fugacity": pandas2ri.ri2py_dataframe(sample.rx2('fugacity')).apply(pd.to_numeric, errors='coerce')})
+                # replace sample fugacity entry with None if there is no fugacity data.
+                if(len(dict_sample_data['fugacity'].index) == 1 and dict_sample_data['fugacity'].index[0] == 'None'):
+                    dict_sample_data['fugacity'] = None
+                else:
+                    dict_sample_data["fugacity"]["fugacity"] = 10**dict_sample_data["fugacity"]["log_fugacity"]
+            
             if get_affinity_energy:
                 dict_sample_data.update({"affinity_energy_raw": pandas2ri.ri2py_dataframe(
                     sample.rx2('affinity_energy_raw')).apply(pd.to_numeric, errors='coerce')})
@@ -1975,6 +2067,9 @@ class AqEquil():
                 {sample_data.names[i]: dict_sample_data})
 
         out_dict.update({"batch_3o": batch_3o})
+        
+        out_dict.update({"water_model":water_model, "grid_temp":grid_temp, "grid_press":grid_press})
+        
         speciation = Speciation(out_dict)
 
         if report_filename != None:
@@ -1993,11 +2088,12 @@ class AqEquil():
 
 
     def create_data0(self,
+                     db,
                      filename,
                      filename_ss=None,
                      data0_formula_ox_name=None,
                      suppress_redox=[],
-                     db="wrm",
+                     water_model="SUPCRT92",
                      exceed_Ttr=True,
                      grid_temps=[0.0100, 50.0000, 100.0000, 150.0000,
                                  200.0000, 250.0000, 300.0000, 350.0000],
@@ -2005,12 +2101,16 @@ class AqEquil():
                      infer_formula_ox=False,
                      generate_template=True,
                      template_name=None,
+                     template_type="strict",
                      verbose=1):
         """
         Create a data0 file from a custom thermodynamic dataset.
         
         Parameters
         ----------
+        db : str
+            Desired three letter code of data0 output.
+            
         filename : str
             Name of csv file containing thermodynamic data in the OBIGT format.
             
@@ -2025,9 +2125,11 @@ class AqEquil():
         suppress_redox : list of str, default []
             Suppress equilibrium between oxidation states of listed elements
             (Cl, H, and O cannot be included).
-        
-        db : str, default "wrm"
-            Desired three letter code of data0 output.
+
+        water_model : str, default "SUPCRT92"
+            This is an experimental feature that is not yet fully supported.
+            Desired water model. Can be either "SUPCRT92", "IAPWS95", or "DEW".
+            These models are described here: http://chnosz.net/manual/water.html
 
         exceed_Ttr : bool, default True
             Calculate Gibbs energies of mineral phases and other species
@@ -2051,6 +2153,8 @@ class AqEquil():
         
         generate_template : bool, default True
             Generate a CSV sample input template customized to this data0?
+            Columns include 'Sample', 'Temperature', 'logfO2', and all strict
+            basis species.
         
         template_name : str, optional
             Name of the sample input template file generated. If no name is
@@ -2058,6 +2162,13 @@ class AqEquil():
             the three letter code given to `db`. Ignored if `generate_template`
             is False.
         
+        template_type : str, either 'strict', 'all basis', or 'all species'
+            Determines which columns are written to the sample template.
+            - 'strict' includes strict basis species
+            - 'all basis' includes strict and auxiliary basis species
+            - 'all species' includes all species in the thermodynamic database
+            Ignored if `generate_template` is False.
+        
         verbose : int, 0, 1, or 2, default 1
             Level determining how many messages are returned during a
             calculation. 2 for all messages, 1 for errors or warnings only,
@@ -2075,12 +2186,62 @@ class AqEquil():
         if self.verbose >= 1:
             print("Creating data0.{}...".format(db), flush=True)
         
+        if len(grid_temps) > 8 or len(grid_temps) < 1:
+            raise Exception("'grid_temps' must have eight values.")
+        if isinstance(grid_press, list):
+            if len(grid_press) > 8 or len(grid_press) < 1:
+                raise Exception("'grid_press' must have eight values.")
+        
         if sum([T >= 10000 for T in grid_temps]):
             raise Exception("Grid temperatures must be below 10000 °C.")
         
         if isinstance(grid_press, list):
-            if sum([T >= 10000 for T in grid_temps]):
+            if sum([P >= 10000 for P in grid_press]):
                 raise Exception("Grid pressures must be below 10000 bars.")
+            
+        if water_model == "SUPCRT92":
+            min_T = 0
+            max_T = 2250
+            min_P = 1
+            max_P = 30000
+        elif water_model == "IAPWS95":
+            min_T = 0
+            max_T = 1000
+            min_P = 1
+            max_P = 10000
+        elif water_model == "DEW":
+            min_T = 0
+            max_T = 1000
+            min_P = 1
+            max_P = 60000
+        else:
+            raise Exception("The water model '{}' ".format(water_model)+"is not "
+                            "recognized. Try 'SUPCRT92', 'IAPWS95', or 'DEW'.")
+        
+        # check that T and P are above minimum values
+        if sum([T <= min_T for T in grid_temps]):
+            print("WARNING: one or more temperatures in 'grid_temps' is below "
+                  "or equal to {} °C".format(min_T)+" and is outside the valid "
+                  "temperature range for the {} water model.".format(water_model))
+        if isinstance(grid_press, list):
+            if sum([P < min_P for P in grid_press]):
+                print("WARNING: one or more pressures in 'grid_press' is below "
+                      "{} bar".format(min_P)+", the minimum valid "
+                      "pressure for the {} water model.".format(water_model))
+        
+        # check that T and P are below maximum values
+        if sum([T > max_T for T in grid_temps]):
+            print("WARNING: one or more temperatures in 'grid_temps' is above "
+                  "{} °C".format(max_T)+", the maximum valid "
+                  "temperature for the {} water model.".format(water_model))
+        if isinstance(grid_press, list):
+            if sum([P > max_P for P in grid_press]):
+                print("WARNING: one or more pressures in 'grid_press' is above "
+                      "{} bar".format(max_P)+", the maximum valid "
+                      "pressure for the {} water model.".format(water_model))
+            
+        if water_model != "SUPCRT92":
+            print("WARNING: water models other than SUPCRT92 are not yet fully supported.")
         
         template = pkg_resources.resource_string(
             __name__, 'data0.min').decode("utf-8")
@@ -2094,6 +2255,9 @@ class AqEquil():
             data0_formula_ox_name = ro.r("NULL")
         if template_name == None:
             template_name = "sample_template_{}.csv".format(db)
+        if template_type not in ['strict', 'all basis', 'all species']:
+            raise Exception("template_type {} ".format(template_type)+"is not"
+                            "recognized. Try 'strict', 'all basis', or 'all species'")
 
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter("always")
@@ -2105,6 +2269,7 @@ class AqEquil():
                                    grid_temps=grid_temps,
                                    grid_press=grid_press,
                                    db=db,
+                                   water_model=water_model,
                                    template=template,
                                    exceed_Ttr=exceed_Ttr,
                                    data0_formula_ox_name=data0_formula_ox_name,
@@ -2112,6 +2277,7 @@ class AqEquil():
                                    infer_formula_ox=infer_formula_ox,
                                    generate_template=generate_template,
                                    template_name=template_name,
+                                   template_type=template_type,
                                    verbose=self.verbose)
     
         for warning in w:
